# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-13 18:42
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_orghierarchy', '0002_organization_replaced_by'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='organization',
            name='responsible_organization',
        ),
        migrations.AddField(
            model_name='organization',
            name='internal_type',
            field=models.CharField(choices=[('normal', 'Normal organization'), ('affiliated', 'Affiliated organization')], default='normal', max_length=20),
        ),
    ]



from django.contrib.auth import get_user_model


def make_admin(username='testadmin', is_superuser=True):
    user_model = get_user_model()
    return user_model.objects.create(username=username, is_staff=True, is_superuser=is_superuser)


def clear_user_perm_cache(user):
    # permission cache, when empty, evaluates to False, but must be deleted anyway
    if getattr(user, '_user_perm_cache', None) is not None:
        delattr(user, '_user_perm_cache')
    if getattr(user, '_perm_cache', None) is not None:
        delattr(user, '_perm_cache')



import factory


class DataSourceFactory(factory.django.DjangoModelFactory):
    id = factory.Sequence(lambda n: "data-source-{0}".format(n))
    name = factory.Faker('text', max_nb_chars=255)

    class Meta:
        model = 'django_orghierarchy.DataSource'


class OrganizationClassFactory(factory.django.DjangoModelFactory):
    data_source = factory.SubFactory(DataSourceFactory)
    origin_id = factory.Faker('text', max_nb_chars=255)
    name = factory.Faker('text', max_nb_chars=255)

    class Meta:
        model = 'django_orghierarchy.OrganizationClass'


class OrganizationFactory(factory.django.DjangoModelFactory):
    data_source = factory.SubFactory(DataSourceFactory)
    origin_id = factory.Faker('text', max_nb_chars=255)
    classification = factory.SubFactory(OrganizationClassFactory)
    name = factory.Faker('text', max_nb_chars=255)

    class Meta:
        model = 'django_orghierarchy.Organization'



from unittest.mock import MagicMock, patch

from django.core.management import call_command, CommandError
from django.test import TestCase

from django_orghierarchy.importers import DataImportError
from django_orghierarchy.models import Organization, OrganizationClass
from django_orghierarchy.utils import get_data_source_model
from .test_importers import mock_request_get


class TestRestImportCommand(TestCase):

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_rest_import_success(self):
        url = 'http://fake.url/organizations/?page=1'
        call_command('import_organizations', url)

        data_source_model = get_data_source_model()
        self.assertEqual(Organization.objects.count(), 3)
        self.assertEqual(data_source_model.objects.count(), 2)
        self.assertEqual(OrganizationClass.objects.count(), 2)

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_rest_import_handle_data_import_error(self):
        url = 'http://not-exist.url/organizatoins/?page=1'
        try:
            call_command('import_organizations', url)
        except DataImportError:
            self.fail('The command does not handle DataImportError')

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_rest_import_with_renamed_source(self):
        url = 'http://fake.url/organizations/?page=1'
        rename_data_source = ['test-source-1: new-source-1', 'test-source-2: new-source-2']
        call_command('import_organizations', url, rename_data_source=rename_data_source)

        data_source_model = get_data_source_model()
        qs = data_source_model.objects.filter(id__in=['new-source-1', 'new-source-2'])
        self.assertTrue(qs.count(), 2)
        self.assertFalse(data_source_model.objects.filter(id__in=['test-source-1', 'test-source-2']).exists())

    def test_rest_import_with_invalid_renamed_source(self):
        url = 'http://fake.url/organizations/?page=1'
        rename_data_source = ['test-source-1: new-source-1', 'test-source-2, new-source-2']
        self.assertRaises(
            CommandError,
            call_command,
            'import_organizations',
            url,
            rename_data_source=rename_data_source,
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-01-09 20:02
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('django_orghierarchy', '0004_add_affiliated_org_permissions'),
    ]

    operations = [
        migrations.AddField(
            model_name='organizationclass',
            name='created_time',
            field=models.DateTimeField(default=django.utils.timezone.now, help_text='The time at which the resource was created'),
        ),
        migrations.AddField(
            model_name='organizationclass',
            name='data_source',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.DJANGO_ORGHIERARCHY_DATASOURCE_MODEL),
        ),
        migrations.AddField(
            model_name='organizationclass',
            name='last_modified_time',
            field=models.DateTimeField(auto_now=True, help_text='The time at which the resource was updated'),
        ),
        migrations.AddField(
            model_name='organizationclass',
            name='origin_id',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AlterField(
            model_name='organization',
            name='created_time',
            field=models.DateTimeField(default=django.utils.timezone.now, help_text='The time at which the resource was created'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='origin_id',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AlterField(
            model_name='organizationclass',
            name='id',
            field=models.CharField(editable=False, max_length=255, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='organizationclass',
            name='name',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterUniqueTogether(
            name='organizationclass',
            unique_together=set([('data_source', 'origin_id')]),
        ),
    ]



import re
from unittest.mock import MagicMock, patch

import requests
from django.test import TestCase

from django_orghierarchy.importers import DataImportError, RestAPIImporter
from django_orghierarchy.models import Organization, OrganizationClass
from django_orghierarchy.utils import get_data_source_model

from .factories import OrganizationFactory


class MockResponse:

    def __init__(self, json_data, status_code=200):
        self.json_data = json_data
        self.status_code = status_code

    def json(self):
        return self.json_data

    def raise_for_status(self):
        if self.status_code != 200:
            raise requests.HTTPError('HTTPError raised')


organization_1 = {
    'id': 111,
    'data_source': 'test-source-1',
    'origin_id': 'ABC-123',
    'classification': 'test-source-1:test-class-1',
    'name': 'Organization-1',
    'founding_date': '2000-01-01',
    'dissolution_date': '2017-01-01',
    'parent': 'http://fake.url/organizations/222/',
    'ignored_field': 'This field will be ignored',
}

organization_2 = {
    'id': 222,
    'data_source': 'test-source-1',
    'origin_id': 'ABC-456',
    'classification': 'test-source-1:test-class-1',
    'name': 'Organization-2',
    'founding_date': '2000-01-01',
    'dissolution_date': None,
    'parent': None,
    'ignored_field': 'This field will be ignored',
}

organization_3 = {
    'id': 333,
    'data_source': 'test-source-2',
    'origin_id': 'XYZ-3',
    'classification': 'test-source-2:test-class-2',
    'name': 'Organizatoin-3',
    'founding_date': '2016-01-01',
    'dissolution_date': None,
    'parent': 'http://fake.url/organizations/111/',
    'ignored_field': 'This field will be ignored',
}

organizations = {
    '111': organization_1,
    '222': organization_2,
    '333': organization_3,
}

tprek_organization_1 = {
    'id': '111',
    'parent_id': '222',
    'organization_type': 'TEST_TYPE_1',
    'name_fi': 'Organization-1',
    'ignored_field': 'This field will be ignored',
}

tprek_organization_2 = {
    'id': '222',
    'organization_type': 'TEST_TYPE_1',
    'name_fi': 'Organization-2',
    'ignored_field': 'This field will be ignored',
}

# TPREK may have missing parent organizations. Pay no heed.
tprek_organization_3 = {
    'id': '333',
    'parent_id': '444',
    'organization_type': 'TEST_TYPE_2',
    'name_fi': 'Organization-3',
    'ignored_field': 'This field will be ignored',
}

tprek_organizations = {
    '111': tprek_organization_1,
    '222': tprek_organization_2,
    '333': tprek_organization_3,
}


def mock_request_get(url):
    m = re.search(r'http://fake.url/organizations/(\d+)/', url)
    if m:
        org_id = m.group(1)
        return MockResponse(organizations[org_id])

    m = re.search(r'http://fake.url/organizations/\?page=(\d+)$', url)
    if m:
        page = m.group(1)
        if page == '1':
            return MockResponse({
                'next': 'http://fake.url/organizations/?page=2',
                'results': [organization_1, organization_2]
            })
        elif page == '2':
            return MockResponse({
                'next': None,
                'results': [organization_3]
            })
    m = re.search(r'http://fake.url/', url)
    if m:
        return MockResponse({'next_page': None, 'items': []}, status_code=200)

    return MockResponse({}, status_code=404)


def mock_tprek_request_get(url):
    m = re.search(r'http://fake.tprek.url/department/', url)
    if m:
        return MockResponse(tprek_organizations.values())

    return MockResponse({}, status_code=404)


class TestRestApiImporter(TestCase):

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def setUp(self):
        self.importer = RestAPIImporter('http://fake.url/organizations/?page=1')

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_custom_config(self):
        config = {
            'next_key': 'next_page',
            'results_key': 'items',
            'fields': ['classification', 'origin_id', 'name', 'parent'],
            'field_config': {
                'classification': {
                    'data_type': 'link',
                }
            }
        }
        importer = RestAPIImporter('http://fake.url/', config)
        self.assertEqual(importer.next_key, 'next_page')
        self.assertEqual(importer.results_key, 'items')

        expected_field_config = {
            'parent': {
                'data_type': 'link',
            },
            'origin_id': {
                'data_type': 'str_lower',
            },
            'classification': {
                'data_type': 'link',
            }
        }
        self.assertDictEqual(importer.field_config, expected_field_config)

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_get_organization_class(self):
        data = {'id': 'test-source:test-class'}
        organization_class = self.importer._get_organization_class(data)
        self.assertEqual(organization_class.id, 'test-source:test-class')
        self.assertEqual(OrganizationClass.objects.count(), 1)

        self.importer._get_organization_class(data)
        self.assertEqual(OrganizationClass.objects.count(), 1)  # fetched from cached

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_get_data_source(self):
        data = {'id': 'test-source'}
        data_source = self.importer._get_data_source(data)
        self.assertEqual(data_source.id, 'test-source')
        data_source_model = get_data_source_model()
        self.assertEqual(data_source_model.objects.count(), 1)

        self.importer._get_data_source(data)
        self.assertEqual(data_source_model.objects.count(), 1)  # fetched from cached

        self.importer.config['rename_data_source'] = {
            'test-source': 'new-source-name'
        }
        data_source = self.importer._get_data_source(data)
        self.assertEqual(data_source.id, 'new-source-name')
        self.assertEqual(data_source_model.objects.count(), 2)

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_import_data(self):
        self.importer.import_data()
        data_source_model = get_data_source_model()
        self.assertEqual(Organization.objects.count(), 3)
        self.assertEqual(data_source_model.objects.count(), 2)
        self.assertEqual(OrganizationClass.objects.count(), 2)

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_import_organization_with_parent(self):
        organization = self.importer._import_organization(organization_1)
        qs = Organization.objects.all()
        # also created parent organization
        self.assertQuerysetEqual(qs, [repr(organization.parent), repr(organization)], ordered=False)
        self.assertEqual(organization.name, 'Organization-1')
        self.assertNotEqual(organization.id, 111)
        self.assertEqual(organization.parent.name, 'Organization-2')
        self.assertNotEqual(organization.parent_id, 222)

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_import_organization_without_parent(self):
        organization = self.importer._import_organization(organization_2)
        qs = Organization.objects.all()
        self.assertQuerysetEqual(qs, [repr(organization)])
        self.assertEqual(organization.name, 'Organization-2')
        self.assertNotEqual(organization.id, 222)

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_import_organization_flip_parents(self):
        organization = self.importer._import_organization(organization_1)
        qs = Organization.objects.all()
        # also created parent organization
        self.assertQuerysetEqual(qs, [repr(organization.parent), repr(organization)], ordered=False)
        self.assertEqual(organization.name, 'Organization-1')
        self.assertNotEqual(organization.id, 111)
        self.assertEqual(organization.parent.name, 'Organization-2')
        self.assertNotEqual(organization.parent_id, 222)
        changed_organization = organization_2.copy()
        new_parent = organization_1.copy()
        changed_organization["parent"] = "http://fake.url/organizations/111/"
        new_parent["parent"] = None
        organizations["111"] = new_parent
        organization = self.importer._import_organization(changed_organization)
        # Now the parents should have switched.
        self.assertQuerysetEqual(qs, [repr(organization.parent), repr(organization)], ordered=False)
        self.assertEqual(organization.name, 'Organization-2')
        self.assertNotEqual(organization.id, 222)
        self.assertEqual(organization.parent.name, 'Organization-1')
        self.assertNotEqual(organization.id, 111)
        # tear down this test, as it had a side effect
        organizations["111"] = organization_1

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_import_organization_update_existing(self):
        organization = OrganizationFactory(
            name='existing-organization',
            origin_id=organization_2['origin_id'],
            data_source=self.importer._import_data_source(organization_2['data_source']),
        )
        self.importer._import_organization(organization_2)
        organization.refresh_from_db()

        self.assertQuerysetEqual(Organization.objects.all(), [repr(organization)])
        self.assertEqual(organization.name, 'Organization-2')

    def test_import_data_source_with_string(self):
        data_source = self.importer._import_data_source('test-data-source')
        data_source_model = get_data_source_model()
        qs = data_source_model.objects.all()
        self.assertQuerysetEqual(qs, [repr(data_source)])
        self.assertEqual(data_source.id, 'test-data-source')

    def test_import_data_source_with_dict_data(self):
        data = {
            'id': 'test-data-source'
        }
        data_source = self.importer._import_data_source(data)
        data_source_model = get_data_source_model()
        qs = data_source_model.objects.all()
        self.assertQuerysetEqual(qs, [repr(data_source)])
        self.assertEqual(data_source.id, 'test-data-source')

    def test_import_organization_class_with_string(self):
        organization_class = self.importer._import_organization_class('test-source-1:test-org-class')
        qs = OrganizationClass.objects.all()
        self.assertQuerysetEqual(qs, [repr(organization_class)])
        self.assertEqual(organization_class.id, 'test-source-1:test-org-class')

    def test_import_organization_class_with_simple_string(self):
        organization_class = self.importer._import_organization_class('class-with-no-source')
        qs = OrganizationClass.objects.all()
        self.assertQuerysetEqual(qs, [repr(organization_class)])
        self.assertEqual(organization_class.id, 'OpenDecisionAPI:class-with-no-source')

    def test_import_organization_class_with_dict_data(self):
        data = {
            'id': 999,
            'origin_id': 'test-org-class',
            'name': 'test-org-class',
            'data_source': 'test-source-1',
        }
        organization_class = self.importer._import_organization_class(data)
        qs = OrganizationClass.objects.all()
        self.assertQuerysetEqual(qs, [repr(organization_class)])
        self.assertEqual(organization_class.name, 'test-org-class')
        self.assertEqual(organization_class.id, 'test-source-1:test-org-class')

    def test_import_organization_class_with_simple_dict_data(self):
        data = {
            'id': 999,
            'name': 'test-org-class'
        }
        organization_class = self.importer._import_organization_class(data)
        qs = OrganizationClass.objects.all()
        self.assertQuerysetEqual(qs, [repr(organization_class)])
        self.assertEqual(organization_class.name, 'test-org-class')
        self.assertEqual(organization_class.id, 'OpenDecisionAPI:999')

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_data_iter(self):
        url = 'http://fake.url/organizations/?page=1'
        iterator = self.importer._data_iter(url)
        iter_len = len(list(iterator))
        self.assertEqual(iter_len, 3)

        url = 'http://fake.url/organizations/?page=2'
        iterator = self.importer._data_iter(url)
        iter_len = len(list(iterator))
        self.assertEqual(iter_len, 1)

        url = 'http://not-exist.url/organizations/'
        self.assertRaises(DataImportError, list, self.importer._data_iter(url))

    def test_get_field_value_empty_value(self):
        value = self.importer._get_field_value(organization_2, 'dissolution_date', {})
        self.assertIsNone(value)

    def test_get_field_value_different_source_field(self):
        config = {
            'source_field': 'id',
        }
        value = self.importer._get_field_value(organization_1, 'origin_id', config)
        self.assertEqual(value, 111)

    def test_get_field_value_not_exist_field(self):
        config = {
            'source_field': 'not-exist-field',
        }
        self.assertRaises(
            DataImportError,
            self.importer._get_field_value,
            organization_1,
            'origin_id',
            config,
        )

    def test_get_field_value_not_exist_data_type(self):
        config = {
            'data_type': 'not-exist-data-type',
        }
        self.assertRaises(
            DataImportError,
            self.importer._get_field_value,
            organization_1,
            'origin_id',
            config,
        )

    @patch('requests.get', MagicMock(return_value=MockResponse('test-get-return')))
    def test_get_field_value_link_data_type(self):
        config = {
            'data_type': 'link',
        }
        value = self.importer._get_field_value({'name': 'http://fake.url/'}, 'name', config)
        self.assertEqual(value, 'test-get-return')

    def test_get_field_value_str_lower_data_type(self):
        config = {
            'data_type': 'str_lower',
        }
        value = self.importer._get_field_value(organization_1, 'origin_id', config)
        self.assertEqual(value, 'abc-123')

    def test_get_field_value_regex_data_type(self):
        config = {
            'data_type': 'regex',
        }
        # test raise DataImportError if no pattern is provided
        self.assertRaises(
            DataImportError,
            self.importer._get_field_value,
            organization_1,
            'origin_id',
            config,
        )

        config = {
            'data_type': 'regex',
            'pattern': r'\w+\-(\d+)',
        }
        value = self.importer._get_field_value(organization_1, 'origin_id', config)
        self.assertEqual(value, '123')

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_get_field_value_related_fields(self):
        value = self.importer._get_field_value(organization_1, 'data_source', {})
        self.assertEqual(value.id, 'test-source-1')

        value = self.importer._get_field_value(organization_1, 'classification', {})
        self.assertEqual(value.id, 'test-source-1:test-class-1')

        value = self.importer._get_field_value(organization_1, 'parent', {'data_type': 'link'})
        self.assertEqual(value.name, 'Organization-2')

    def test_get_regex_data(self):
        pattern = r'\w+\-(\d+)'
        self.assertRaises(
            DataImportError,
            self.importer._get_regex_data,
            '123',
            pattern,
        )

        data = self.importer._get_regex_data('abc-123', pattern)
        self.assertEqual(data, '123')

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_get_link_data(self):
        invalid_url = 'abc.123'
        self.assertRaises(DataImportError, self.importer._get_link_data, invalid_url)

        url = 'http://not-exist.url/organizations/111/'
        self.assertRaises(DataImportError, self.importer._get_link_data, url)

        url = 'http://fake.url/organizations/111/'
        data = self.importer._get_link_data(url)
        self.assertEqual(data, organization_1)


class TestTprekRestApiImporter(TestRestApiImporter):
    # here we test all the features of the TPREK import that are different from default REST import

    @patch('requests.get', MagicMock(side_effect=mock_tprek_request_get))
    def setUp(self):
        self.config = {
            'next_key': None,
            'results_key': None,
            'fields': [
                'origin_id', 'classification',
                'name', 'parent',
            ],
            'update_fields': [
                'classification', 'name', 'parent',
            ],
            'field_config': {
                'parent': {
                    'source_field': 'parent_id',
                    'data_type': 'org_id',
                    'optional': True,
                },
                'origin_id': {
                    'source_field': 'id',
                    'data_type': 'value',
                },
                'classification': {
                    'source_field': 'organization_type',
                    'optional': True,
                },
                'name': {
                    'source_field': 'name_fi',
                    'optional': True,
                }
            },
            'default_data_source': 'tprek',
            'default_parent_organization': 'Pääkaupunkiseudun toimipisterekisteri'
        }
        self.importer = RestAPIImporter('http://fake.tprek.url/department/', self.config)

    @patch('requests.get', MagicMock(side_effect=mock_tprek_request_get))
    def test_get_data_source(self):
        data = {'id': 'tprek'}
        data_source = self.importer._get_data_source(data)
        self.assertEqual(data_source.id, 'tprek')
        data_source_model = get_data_source_model()
        self.assertEqual(data_source_model.objects.count(), 1)

        self.importer._get_data_source(data)
        self.assertEqual(data_source_model.objects.count(), 1)  # fetched from cached

        self.importer.config['rename_data_source'] = {
            'tprek': 'new-source-name'
        }
        data_source = self.importer._get_data_source(data)
        self.assertEqual(data_source.id, 'new-source-name')
        self.assertEqual(data_source_model.objects.count(), 2)

    @patch('requests.get', MagicMock(side_effect=mock_tprek_request_get))
    def test_get_field_value_related_fields(self):
        value = self.importer._get_field_value({'data_source': 'tprek'}, 'data_source', {})
        self.assertEqual(value.id, 'tprek')

        value = self.importer._get_field_value(tprek_organization_1, 'classification',
                                               self.config['field_config']['classification'])
        self.assertEqual(value.id, 'tprek:TEST_TYPE_1')

        value = self.importer._get_field_value(tprek_organization_1, 'parent',
                                               self.config['field_config']['parent'])
        self.assertEqual(value.id, 'tprek:222')

    @patch('requests.get', MagicMock(side_effect=mock_tprek_request_get))
    def test_data_iter(self):
        url = 'http://fake.tprek.url/department/'
        iterator = self.importer._data_iter(url)
        iter_len = len(list(iterator))
        self.assertEqual(iter_len, 3)

        url = 'http://not-exist.url/organizations/'
        self.assertRaises(DataImportError, list, self.importer._data_iter(url))

    @patch('requests.get', MagicMock(side_effect=mock_tprek_request_get))
    def test_import_data(self):
        self.importer.import_data()
        data_source_model = get_data_source_model()
        self.assertEqual(Organization.objects.count(), 4)
        self.assertEqual(data_source_model.objects.count(), 1)
        self.assertEqual(OrganizationClass.objects.count(), 2)

    def test_import_organization_class_with_simple_string(self):
        organization_class = self.importer._import_organization_class('class-with-no-source')
        qs = OrganizationClass.objects.all()
        self.assertQuerysetEqual(qs, [repr(organization_class)])
        self.assertEqual(organization_class.id, 'tprek:class-with-no-source')

    def test_import_organization_class_with_dict_data(self):
        pass

    def test_import_organization_class_with_simple_dict_data(self):
        pass

    def test_import_data_source_with_string(self):
        pass

    def test_import_data_source_with_dict_data(self):
        pass

    def test_import_organization_update_existing(self):
        organization = OrganizationFactory(
            name='existing-organization',
            origin_id=tprek_organization_2['id'],
            data_source=self.importer._import_data_source('tprek'),
        )
        self.importer._import_organization(tprek_organization_2)
        organization.refresh_from_db()
        default_parent = Organization.objects.get(id='tprek:tprek')

        self.assertQuerysetEqual(Organization.objects.all(), [repr(default_parent), repr(organization)], ordered=False)
        self.assertEqual(organization.name, 'Organization-2')

    @patch('requests.get', MagicMock(side_effect=mock_tprek_request_get))
    def test_import_organization_with_parent(self):
        organization = self.importer._import_organization(tprek_organization_1)
        qs = Organization.objects.all()
        default_parent = Organization.objects.get(id='tprek:tprek')

        # also created parent organization.
        # parent will have the default parent organization
        self.assertQuerysetEqual(qs, [repr(default_parent), repr(organization.parent), repr(organization)], ordered=False)
        self.assertEqual(organization.name, 'Organization-1')
        self.assertEqual(organization.id, 'tprek:111')
        self.assertEqual(organization.parent.name, 'Organization-2')
        self.assertEqual(organization.parent_id, 'tprek:222')
        self.assertEqual(organization.parent.parent.name, 'Pääkaupunkiseudun toimipisterekisteri')
        self.assertEqual(organization.parent.parent_id, 'tprek:tprek')

    @patch('requests.get', MagicMock(side_effect=mock_request_get))
    def test_import_organization_with_missing_parent(self):
        organization = self.importer._import_organization(tprek_organization_3)
        print(organization)
        qs = Organization.objects.all()

        # also created parent organization.
        # parent was not found, so organization will have the default
        default_parent = Organization.objects.get(id='tprek:tprek')
        self.assertQuerysetEqual(qs, [repr(default_parent), repr(organization)])
        self.assertEqual(organization.name, 'Organization-3')
        self.assertEqual(organization.id, 'tprek:333')
        self.assertEqual(organization.parent.name, 'Pääkaupunkiseudun toimipisterekisteri')
        self.assertEqual(organization.parent_id, 'tprek:tprek')

    @patch('requests.get', MagicMock(side_effect=mock_tprek_request_get))
    def test_import_organization_without_parent(self):
        organization = self.importer._import_organization(tprek_organization_2)
        qs = Organization.objects.all()
        default_parent = Organization.objects.get(id='tprek:tprek')
        self.assertQuerysetEqual(qs, [repr(default_parent), repr(organization)], ordered=False)
        self.assertEqual(organization.name, 'Organization-2')
        self.assertEqual(organization.id, 'tprek:222')
        # organization will have the default parent organization
        self.assertEqual(organization.parent.name, 'Pääkaupunkiseudun toimipisterekisteri')
        self.assertEqual(organization.parent_id, 'tprek:tprek')

    @patch('requests.get', MagicMock(side_effect=mock_tprek_request_get))
    def test_import_organization_flip_parents(self):
        organization = self.importer._import_organization(tprek_organization_1)
        qs = Organization.objects.all()
        default_parent = Organization.objects.get(id='tprek:tprek')

        # also created parent organization.
        # parent will have the default parent organization
        self.assertQuerysetEqual(qs, [repr(default_parent), repr(organization.parent), repr(organization)], ordered=False)
        self.assertEqual(organization.name, 'Organization-1')
        self.assertEqual(organization.id, 'tprek:111')
        self.assertEqual(organization.parent.name, 'Organization-2')
        self.assertEqual(organization.parent_id, 'tprek:222')
        self.assertEqual(organization.parent.parent.name, 'Pääkaupunkiseudun toimipisterekisteri')
        self.assertEqual(organization.parent.parent_id, 'tprek:tprek')

        changed_organization = tprek_organization_2.copy()
        new_parent = tprek_organization_1.copy()
        changed_organization["parent_id"] = "111"
        new_parent["parent_id"] = None
        tprek_organizations["111"] = new_parent

        # We must init a new importer, since importing by organization id will use a cached dict of all the
        # organizations in the importer.
        new_importer = RestAPIImporter('http://fake.tprek.url/department/', self.config)
        organization = new_importer._import_organization(changed_organization)
        # Now the parents should have switched.
        self.assertQuerysetEqual(qs, [repr(default_parent), repr(organization.parent), repr(organization)], ordered=False)
        self.assertEqual(organization.name, 'Organization-2')
        self.assertEqual(organization.id, 'tprek:222')
        self.assertEqual(organization.parent.name, 'Organization-1')
        self.assertEqual(organization.parent_id, 'tprek:111')
        self.assertEqual(organization.parent.parent.name, 'Pääkaupunkiseudun toimipisterekisteri')
        self.assertEqual(organization.parent.parent_id, 'tprek:tprek')
        # tear down this test, as it had a side effect
        tprek_organizations["111"] = tprek_organization_1



from django.db import models

from django_orghierarchy.models import AbstractDataSource


class CustomDataSource(AbstractDataSource):
    extra_field = models.CharField(max_length=200)


class CustomPrimaryKeyDataSource(models.Model):
    id = models.CharField(max_length=100, primary_key=True)
    name = models.CharField(max_length=255)



from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase

from .factories import DataSourceFactory, OrganizationFactory


class OrganizationAPITestCase(APITestCase):

    def setUp(self):
        data_source = DataSourceFactory(name='abc')
        self.organization = OrganizationFactory(data_source=data_source, origin_id='123')

    def test_organization_list(self):
        url = reverse('api:organization-list')
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_organization(self):
        url = reverse('api:organization-detail', kwargs={'pk': self.organization.pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_modify_organization(self):
        url = reverse('api:organization-detail', kwargs={'pk': self.organization.pk})
        response = self.client.post(url, {})
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_create_organization(self):
        url = reverse('api:organization-list')
        response = self.client.post(url, {})
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)



import copy
import logging
import re
from enum import Enum

import requests
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator
from django.db import transaction

from .models import OrganizationClass, Organization
from .utils import get_data_source_model

logger = logging.getLogger(__name__)


class DataImportError(Exception):
    pass


class DataType(Enum):
    VALUE = 'value'
    STR_LOWER = 'str_lower'
    LINK = 'link'
    REGEX = 'regex'
    ORG_ID = 'org_id'


class RestAPIImporter:
    """This class allows importing organization data from a REST API endpoint. The default
    configuration supports the 6aika Open Decision API specification:
    https://github.com/6aika/api-paatos

    The importer will also create data sources and organization classes if does not exist.
    If a single value is provided to `data_source` and `classification` fields, then the
    importer assumes the value is the name of the data source or organization class. If
    an object (dict) is provided to `data_source` and `classification` fields, then the
    importer will create objects with object data. Either full object data or a link to
    full object data should be provided to parent field so that the importer can create
    parent organization before creating child organization.

    For use in multi-source setups, the importer may optionally create a parent organization
    for the data source that is the common parent of all imported top-level organizations.
    This way, separate organization hierarchies from different data sources may be stored
    in the same application.

    The client code can override the default config or pass in a config object to allow
    the importer parsing different REST API structures.

    Config options:
        - next_key: The url to the next page if the source data is paginated, or None if not.
        - results_key: The object key to the list of organization objects, or None if the list is at
            root level.
        - fields: The fields of which the values will be imported. The same fields will be imported
            in organization classes, if present.
        - update_fields: The fields to update if the organization with same origin_id and data
            source exists.
        - field_config: Configs for each field. Config options:
            - source_field: The source data object key where the field value comes from.
                Defaults to original field name.
            - data_type: The data type for the field, can be 'value', 'str_lower', 'link', 'regex' or 'org_id'.
                Defaults to 'value'. If the data type is 'value', it will return the value
                from source data; if the data type is 'str_lower', it will return the value stringified
                and lowercased; if the data type is 'link', it will return the data fetched
                from the link; if the data type is 'regex', it will return the value extracted
                from the given pattern. If the data type is 'org_id', it will return the organization
                in the JSON with the given 'id'.
            - optional: Whether the importer will continue if the field is missing.
                Defaults to False.
        - rename_data_source: Data sources that are renamed during import.
        - default_data_source: Data source id to use for objects without data source.
        - default_parent_organization: If set, name of the parent organization to be created for all
            imported top level organizations.

    Example config:
        {
            'next_key': 'next',
            'results_key: 'results',
            'fields': ['origin_id', 'data_source', 'classification'],
            'update_fields': ['classification'],
            'field_config': {
                'origin_id': {
                    'source_field': 'id',
                    'data_type': 'regex',
                    'pattern': r'abc:(\d+)',
                },
                'parent': {
                    'data_type': 'link',
                }
            },
            'rename_data_source': {
                'original_name_1': 'new_name_1',
                'original_name_2': 'new_name_2',
            },
            'default_data_source': 'new_name_1',
            'default_parent_organization': None
        }
    """

    paatos_config = {
        'next_key': 'next',
        'results_key': 'results',
        'fields': [
            'data_source', 'origin_id', 'classification',
            'name', 'founding_date', 'dissolution_date',
            'parent',
        ],
        'update_fields': [
            'classification', 'name', 'founding_date',
            'dissolution_date', 'parent',
        ],
        'field_config': {
            'parent': {
                'data_type': 'link',
            },
            'origin_id': {
                'data_type': 'str_lower',
            }
        },
        'default_data_source': 'OpenDecisionAPI'
    }

    tprek_config = {
        'next_key': None,
        'results_key': None,
        'fields': [
            'origin_id', 'classification',
            'name', 'parent',
        ],
        'update_fields': [
            'classification', 'name', 'parent',
        ],
        'field_config': {
            'parent': {
                'source_field': 'parent_id',
                'data_type': 'org_id',
                'optional': True,
            },
            'origin_id': {
                'source_field': 'id',
                'data_type': 'value',
            },
            'classification': {
                'source_field': 'organization_type',
                'optional': True,
            },
            'name': {
                'source_field': 'name_fi',
                'optional': True,
            }
        },
        'default_data_source': 'tprek',
        'default_parent_organization': 'Pääkaupunkiseudun toimipisterekisteri'
    }

    default_config = paatos_config

    def __init__(self, url, config=None):
        self.url = url

        self.config = copy.deepcopy(self.default_config)
        config = copy.deepcopy(config)
        if config:
            # only consider fields listed in given config when merging field configs
            default_field_config = self.config.pop('field_config')
            given_field_config = config.pop('field_config')
            merged_field_config = {}
            for field in config.get('fields', None):
                if field in given_field_config:
                    merged_field_config[field] = given_field_config[field]
                elif field in default_field_config:
                    merged_field_config[field] = default_field_config[field]
            self.config['field_config'] = merged_field_config
            self.config.update(config)
        logger.info('Importing organization data from %s with the following config:' % self.url)
        logger.info(self.config)
        self.related_import_methods = {
            'data_source': self._import_data_source,
            'classification': self._import_organization_class,
            'parent': self._import_organization,
        }

        self._organization_classes = {}
        self._data_sources = {}
        self._default_parent = None
        if self.config.get('default_parent_organization', None):
            origin_id_config = self.config['field_config'].get('origin_id', None)
            origin_id_field = origin_id_config['source_field'] if origin_id_config else 'origin_id'
            data_source_config = self.config['field_config'].get('data_source', None)
            data_source_field = data_source_config['source_field'] if data_source_config else 'data_source'
            name_config = self.config['field_config'].get('name', None)
            name_field = name_config['source_field'] if name_config else 'name'
            default_parent_data = {
                origin_id_field: self.config['default_data_source'],
                data_source_field: self.config['default_data_source'],
                name_field: self.config['default_parent_organization']
            }
            self._default_parent = self._import_organization(default_parent_data)
        # Save all the data from the REST endpoint in a single dict.
        # This is for cases where we need to access the data by id instead of link.
        # Caching is necessary for importing all parent organizations before their children.
        self._data_dict = {}
        for data_item in self._data_iter(self.url):
            self._data_dict[data_item["id"]] = data_item

    @property
    def fields(self):
        return self.config['fields']

    @property
    def update_fields(self):
        return self.config['update_fields']

    @property
    def field_config(self):
        return self.config['field_config']

    @property
    def next_key(self):
        """Object key that stores the link to the next page"""
        return self.config['next_key']

    @property
    def results_key(self):
        """Object key that stores the list of organizations"""
        return self.config['results_key']

    @property
    def rename_data_source(self):
        """Data source renaming configs"""
        return self.config.get('rename_data_source') or {}

    @property
    def default_data_source(self):
        """Default data source string"""
        return self.config['default_data_source']

    def _get_organization_class(self, data):
        """Get organization class for the given object data

        The method will first try to get the organization class from cache, and
        then get from database if not cached.
        """
        # organization class supports id, data_source, origin_id and name.
        supported_fields = {'id', 'origin_id', 'data_source', 'name'}
        identifier = data.get('id')
        # organization class requires data source and origin_id.
        if isinstance(identifier, str) and ':' in identifier:
            if 'data_source' not in data:
                data['data_source'] = identifier.split(':')[0]
            if 'origin_id' not in data:
                data['origin_id'] = identifier.split(':')[1]
        # provided id used if origin id missing
        if 'origin_id' not in data or not data['origin_id']:
            data['origin_id'] = identifier
        # default data source used if missing
        if 'data_source' not in data or not data['data_source']:
            data['data_source'] = self.default_data_source
        # reformat id to fit our model
        if not isinstance(identifier, str) or ':' not in identifier:
            data['id'] = data['data_source'] + ':' + str(identifier)
        data['data_source'] = self.related_import_methods['data_source'](data['data_source'])
        # extra fields should not crash the import. Only use specified fields.
        data = {field: value for (field, value) in data.items() if field in supported_fields}
        if identifier not in self._organization_classes:
            organization_class, _ = OrganizationClass.objects.get_or_create(**data)
            self._organization_classes[identifier] = organization_class
        return self._organization_classes[identifier]

    def _get_data_source(self, data):
        """Get data source for the given object data

        The method will first try to get the data source from cache, and
        then get from database if not cached.
        """
        identifier = data['id']

        if identifier in self.rename_data_source:
            identifier = self.rename_data_source[identifier]
            data['id'] = identifier

        if identifier not in self._data_sources:
            data_source_model = get_data_source_model()
            data_source, _ = data_source_model.objects.get_or_create(**data)
            self._data_sources[identifier] = data_source
        return self._data_sources[identifier]

    def import_data(self):
        """Import data"""
        for data_item in self._data_dict.values():
            self._import_organization(data_item)

    @transaction.atomic
    def _import_organization(self, data):
        """Import single organization.

        The organization import is transactional so, for example, we do not
        accidentally save organization without parent if the parent organization
        saving failed for some reason.
        """
        # id must always be present
        config = self.field_config.get('origin_id') or {}
        if isinstance(data, dict):
            incoming_data = data
        else:
            raise DataImportError('Organization data must be dict.')
        origin_id = self._get_field_value(incoming_data, 'origin_id', config)

        # data source may be missing altogether, or it may be optional
        data_source = None
        config = self.field_config.get('data_source') or {}
        try:
            data_source = self._get_field_value(incoming_data, 'data_source', config)
        except DataImportError as exception:
            if not config or config.get('optional'):
                pass
            else:
                raise exception
        if not data_source:
            data_source = self._get_field_value(
                {'data_source': self.default_data_source}, 'data_source', {'data_type': 'value'})

        # id is never imported in default config

        try:
            values_to_update = {}
            for field in self.update_fields:
                config = self.field_config.get(field) or {}
                try:
                    value = self._get_field_value(incoming_data, field, config)
                except DataImportError as exception:
                    if config.get('optional'):
                        continue
                    else:
                        raise exception
                values_to_update[field] = value

            # Organization parent (and its parent) have been imported and updated recursively. If one of
            # them used to be the descendant of this organization, this might result in mptt InvalidMove
            # exception if this organization instance is not up to date with the database.
            # See https://github.com/django-mptt/django-mptt/issues/650

            # Therefore, we may only fetch the organization from the db *after* all its parents have been
            # processed, to get up to date status of the mptt tree before saving each organization.
            # enforce lower case id standard, but recognize upper case ids as equal:
            organization = Organization.objects.get(origin_id__iexact=origin_id, data_source=data_source)
            logger.info('Organization already exists: {0}'.format(organization.id))
            for field, value in values_to_update.items():
                setattr(organization, field, value)
            if (
                self.config.get('default_parent_organization', None)
                and not organization.parent
                and organization != self._default_parent
            ):
                organization.parent = self._default_parent
            organization.save()

            return organization
        except Organization.DoesNotExist:
            object_data = {'origin_id': origin_id, 'data_source': data_source}
            for field in self.fields:
                config = self.field_config.get(field) or {}
                try:
                    object_data[field] = self._get_field_value(incoming_data, field, config)
                except DataImportError as exception:
                    if config.get('optional'):
                        continue
                    else:
                        raise exception
            organization = Organization.objects.create(**object_data)
            if (
                self.config.get('default_parent_organization', None)
                and not organization.parent
                and organization != self._default_parent
            ):
                organization.parent = self._default_parent
                organization.save()
            return organization

    def _import_data_source(self, data):
        """Import data source

        If a single value is provided to data_source field, it assume it's
        the id of the data source.
        """
        if isinstance(data, dict):
            object_data = data
        else:
            object_data = {'id': data}

        return self._get_data_source(object_data)

    def _import_organization_class(self, data):
        """Import organization class.

        If a single value is provided to classification field, it assume it's
        the id of the organization class.

        If a dict is provided, origin_id is used if present, otherwise id.
        """
        if isinstance(data, dict):
            object_data = {k: v for k, v in data.items() if k != 'id'}
            object_data['id'] = data['origin_id'] if 'origin_id' in data else data['id']
        else:
            object_data = {'id': data}

        return self._get_organization_class(object_data)

    def _data_iter(self, url):
        """Iterate over data items in the REST API endpoint.

        The iterator will follow over next page links if available.
        """
        logger.info('Start reading data from {0} ...'.format(url))

        r = requests.get(url)
        try:
            r.raise_for_status()
        except requests.HTTPError as e:
            raise DataImportError(e)

        data = r.json()
        for data_item in data[self.results_key] if self.results_key else data:
            yield data_item

        logger.info('Reading data from {0} completed'.format(url))

        if self.next_key and data[self.next_key]:
            next_url = data[self.next_key]
            yield from self._data_iter(next_url)


    def _get_field_value(self, data_item, field, config):
        """Get source value for the field from the data item

        If source_field is specified in field config, the source_field
        will be used to get the source value, otherwise the model field
        will be used.

        If the data type is DataType.VALUE, the value will be returned;
        If the data type is DataType.STR_LOWER, the value will be returned stringified to lower case;
        If the data type is DataType.ORG_ID, corresponding organization will be returned;
        If the data type is DataType.LINK, the data in the link will be returned;
        If the data type is DataType.REGEX, the extracted value for the pattern will be returned.
        """
        source_field = config.get('source_field') or field
        try:
            value = data_item[source_field]
        except KeyError:
            raise DataImportError('Field not found in source data: {0}'.format(source_field))

        if not value:
            return value

        if config.get('data_type'):
            try:
                data_type = DataType(config['data_type'])
            except ValueError:
                raise DataImportError('Invalid data type: {0}. Supported data types are: {1}'.format(
                    config['data_type'],
                    ', '.join([e.value for e in DataType]),
                ))
        else:
            data_type = DataType.VALUE

        if data_type == DataType.STR_LOWER:
            value = str(value).lower()
        elif data_type == DataType.ORG_ID:
            value = self._data_dict.get(value, None)
        elif data_type == DataType.LINK:
            value = self._get_link_data(value)
        elif data_type == DataType.REGEX:
            try:
                pattern = config['pattern']
            except KeyError:
                raise DataImportError('No regex pattern provided for the field: {0}'.format(field))
            value = self._get_regex_data(value, pattern)

        # import related objects
        if field in self.related_import_methods:
            import_method = self.related_import_methods[field]
            value = import_method(value)

        return value

    @staticmethod
    def _get_regex_data(value, pattern):
        """Extract value from original string value with the given regex pattern"""
        match = re.search(pattern, value)
        if match:
            data = match.group(1)
        else:
            raise DataImportError(
                'Cannot extract value from string {0} with pattern {1}'.format(value, pattern)
            )
        return data

    @staticmethod
    def _get_link_data(value):
        """Get data fetched from the link"""
        validator = URLValidator()
        try:
            validator(value)
        except ValidationError:
            raise DataImportError('Invalid URL: {0}'.format(value))

        r = requests.get(value)

        try:
            r.raise_for_status()
        except requests.HTTPError as e:
            raise DataImportError(e)

        return r.json()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-15 09:14
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_orghierarchy', '0003_rm_responsible_org_add_internal_type'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='organization',
            options={'permissions': (('add_affiliated_organization', 'Can add affiliated organization'), ('change_affiliated_organization', 'Can change affiliated organization'), ('delete_affiliated_organization', 'Can delete affiliated organization'))},
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-28 13:03
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_orghierarchy', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='organization',
            name='replaced_by',
            field=models.OneToOneField(blank=True, help_text='The organization that replaces this organization', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='replaced_organization', to='django_orghierarchy.Organization'),
        ),
    ]



import copy

from django.core.management.base import BaseCommand, CommandError

from django_orghierarchy.importers import DataImportError, RestAPIImporter


class Command(BaseCommand):
    help = 'Import organization data from a REST API'

    def add_arguments(self, parser):
        parser.add_argument(
            'url',
            metavar='URL',
            help='REST API endpoint from where the organization data will be imported',
        )

        parser.add_argument(
            '-s',
            '--rename-data-source',
            dest='rename_data_source',
            nargs='+',
            default=['helsinki:ahjo'],
            help='Rename data sources. Renaming should be specified as <old_identifier>:<new_identifier>'
        )

        parser.add_argument(
            '-c',
            '--config',
            dest='config',
            default='paatos',
            choices=['paatos', 'tprek'],
            help='REST API configuration to use when importing.'
        )

    def _parse_rename_data_source(self, value):
        try:
            old_name, new_name = value.split(':')
        except ValueError:
            raise CommandError(
                'Invalid data source renaming. Renaming should be specified as <old_identifier>:<new_identifier>'
            )

        return old_name.strip(), new_name.strip()

    def handle(self, *args, **options):
        url = options['url']

        config = {}
        if options['config']:
            config = copy.deepcopy(getattr(RestAPIImporter, '%s_config' % options['config']))
        if options['rename_data_source']:
            rename_data_source = {}
            for item in options['rename_data_source']:
                old_name, new_name = self._parse_rename_data_source(item)
                rename_data_source[old_name] = new_name
            config['rename_data_source'] = rename_data_source

        importer = RestAPIImporter(url, config)
        try:
            importer.import_data()
        except DataImportError as e:
            self.stderr.write('Import failed: {0}'.format(e))
        else:
            self.stdout.write('Import completed successfully')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-23 15:13
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import mptt.fields
import swapper


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        swapper.dependency('django_orghierarchy', 'DataSource')
    ]

    operations = [
        migrations.CreateModel(
            name='DataSource',
            fields=[
                ('id', models.CharField(max_length=100, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'swappable': swapper.swappable_setting('django_orghierarchy', 'DataSource'),
            },
        ),
        migrations.CreateModel(
            name='Organization',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('origin_id', models.CharField(max_length=255, unique=True)),
                ('name', models.CharField(help_text='A primary name, e.g. a legally recognized name', max_length=255)),
                ('founding_date', models.DateField(blank=True, help_text='A date of founding', null=True)),
                ('dissolution_date', models.DateField(blank=True, help_text='A date of dissolution', null=True)),
                ('created_time', models.DateTimeField(auto_now_add=True, help_text='The time at which the resource was created')),
                ('last_modified_time', models.DateTimeField(auto_now=True, help_text='The time at which the resource was updated')),
                ('lft', models.PositiveIntegerField(db_index=True, editable=False)),
                ('rght', models.PositiveIntegerField(db_index=True, editable=False)),
                ('tree_id', models.PositiveIntegerField(db_index=True, editable=False)),
                ('level', models.PositiveIntegerField(db_index=True, editable=False)),
            ],
        ),
        migrations.CreateModel(
            name='OrganizationClass',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
            ],
            options={
                'verbose_name_plural': 'Organization classes',
                'verbose_name': 'Organization class',
            },
        ),
        migrations.AddField(
            model_name='organization',
            name='classification',
            field=models.ForeignKey(blank=True, help_text='An organization category, e.g. committee', null=True, on_delete=django.db.models.deletion.PROTECT, to='django_orghierarchy.OrganizationClass'),
        ),
        migrations.AddField(
            model_name='organization',
            name='created_by',
            field=models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='created_organizations', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='organization',
            name='data_source',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=swapper.get_model_name('django_orghierarchy', 'DataSource')),
        ),
        migrations.AddField(
            model_name='organization',
            name='last_modified_by',
            field=models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='modified_organizations', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='organization',
            name='parent',
            field=mptt.fields.TreeForeignKey(blank=True, help_text='The organizations that contain this organization', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='django_orghierarchy.Organization'),
        ),
        migrations.AddField(
            model_name='organization',
            name='responsible_organization',
            field=models.ForeignKey(blank=True, help_text='Responsible organization', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='affiliated_organization', to='django_orghierarchy.Organization'),
        ),
        migrations.AlterUniqueTogether(
            name='organization',
            unique_together=set([('data_source', 'origin_id')]),
        ),
        migrations.AddField(
            model_name='organization',
            name='admin_users',
            field=models.ManyToManyField(blank=True, related_name='admin_organizations', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='organization',
            name='regular_users',
            field=models.ManyToManyField(blank=True, related_name='organization_memberships', to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 3.1.2 on 2021-03-17 15:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_orghierarchy', '0009_add_change_organization_regular_users_permission'),
    ]

    operations = [
        migrations.RenameField(
            model_name='datasource',
            old_name='user_editable',
            new_name='user_editable_organizations',
        ),
    ]



from django import forms
from django.forms import widgets
from django.utils.translation import ugettext as _

from .models import Organization
from .utils import get_data_source_model


class OrganizationForm(forms.ModelForm):

    class Meta:
        model = Organization
        fields = (
            'data_source', 'origin_id', 'classification',
            'name', 'founding_date', 'dissolution_date',
            'internal_type', 'parent', 'admin_users',
            'regular_users', 'replaced_by',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # the fields can be dynamically exclude, for example set them to readonly in admin
        if 'replaced_by' in self.fields:
            # An organization cannot be replaced by an organization that have already been replaced
            qs = Organization.objects.filter(replaced_by__isnull=True)
            if self.instance.id:
                # prevent self reference
                qs = qs.exclude(id=self.instance.id)
            self.fields['replaced_by'].queryset = qs

        if 'data_source' in self.fields:
            # Only allow selecting data source within editable sources
            self.fields['data_source'].queryset = get_data_source_model().objects.filter(user_editable_organizations=True)

        if 'parent' in self.fields and self.instance.id:
            # prevent recursive reference
            desc_ids = self.instance.get_descendants(include_self=True).values_list('id', flat=True)
            self.fields['parent'].queryset = Organization.objects.exclude(id__in=desc_ids)

    def clean(self):
        cleaned_data = super().clean()
        internal_type = cleaned_data.get('internal_type')
        parent = cleaned_data.get('parent')

        if internal_type == Organization.AFFILIATED and parent is None:
            raise forms.ValidationError(_('Affiliated organization must have a parent organization'))

        return cleaned_data


class SubOrganizationForm(forms.ModelForm):
    default_internal_type = Organization.NORMAL

    class Meta:
        model = Organization
        fields = (
            'name', 'founding_date', 'classification', 'data_source', 'origin_id', 'internal_type'
        )
        widgets = {
            'internal_type': widgets.HiddenInput,
        }

    def __init__(self, *args, **kwargs):
        if 'initial' not in kwargs:
            kwargs['initial'] = {}
        kwargs['initial']['internal_type'] = self.default_internal_type
        super().__init__(*args, **kwargs)

        # the fields can be dynamically exclude, for example set them to readonly in admin
        if 'data_source' in self.fields:
            # Only allow selecting data source within editable sources
            self.fields['data_source'].queryset = get_data_source_model().objects.filter(user_editable_organizations=True)

    def clean_internal_type(self):
        return self.initial['internal_type']  # do not allow changing internal_type


class AffiliatedOrganizationForm(SubOrganizationForm):
    default_internal_type = Organization.AFFILIATED



from django.test import TestCase

from django_orghierarchy.forms import OrganizationForm, SubOrganizationForm, AffiliatedOrganizationForm
from django_orghierarchy.models import Organization
from .factories import OrganizationClassFactory, DataSourceFactory, OrganizationFactory


class TestOrganizationForm(TestCase):

    def setUp(self):
        self.organization = OrganizationFactory()
        self.data_source = DataSourceFactory(user_editable_organizations=True)

    def test_init_without_instance(self):
        form = OrganizationForm()
        self.assertQuerysetEqual(form.fields['parent'].queryset, [repr(self.organization)])

    def test_init_with_instance(self):
        form = OrganizationForm(instance=self.organization)
        self.assertQuerysetEqual(form.fields['parent'].queryset, [])

    def test_replaced_by_field_queryset_exclude_already_replaced(self):
        OrganizationFactory(replaced_by=self.organization)
        organization_2 = OrganizationFactory()

        form = OrganizationForm()
        queryset = form.fields['replaced_by'].queryset
        self.assertQuerysetEqual(queryset, [repr(self.organization), repr(organization_2)], ordered=False)

        form = OrganizationForm(instance=organization_2)
        queryset = form.fields['replaced_by'].queryset
        self.assertQuerysetEqual(queryset, [repr(self.organization)])

    def test_clean(self):
        form_data = {
            'data_source': self.data_source.id,
            'origin_id': 'abc',
            'classification': self.organization.classification_id,
            'name': 'test org',
            'internal_type': Organization.AFFILIATED,
            'parent': '',
        }
        form = OrganizationForm(form_data)
        self.assertFalse(form.is_valid())

        form_data['parent'] = self.organization.id
        form = OrganizationForm(form_data)
        self.assertTrue(form.is_valid())


class TestSubOrganizationForm(TestCase):

    def setUp(self):
        self.data_source = DataSourceFactory(user_editable_organizations=True)
        self.organization_class = OrganizationClassFactory()

    def test__init__(self):
        form = SubOrganizationForm()
        self.assertEqual(form.initial['internal_type'], Organization.NORMAL)

        form = SubOrganizationForm(initial={'internal_type': Organization.AFFILIATED})
        self.assertEqual(form.initial['internal_type'], Organization.NORMAL)

    def test_clean_internal_type(self):
        form_data = {
            'internal_type': Organization.NORMAL,
            'data_source': self.data_source.id,
            'origin_id': 'sub-org',
            'classification': self.organization_class.id,
            'name': 'sub organization',
        }
        form = SubOrganizationForm(form_data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['internal_type'], Organization.NORMAL)

        form_data['internal_type'] = Organization.AFFILIATED
        form = SubOrganizationForm(form_data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['internal_type'], Organization.NORMAL)


class TestAffiliatedOrganizationForm(TestCase):

    def setUp(self):
        self.data_source = DataSourceFactory(user_editable_organizations=True)
        self.organization_class = OrganizationClassFactory()

    def test__init__(self):
        form = AffiliatedOrganizationForm()
        self.assertEqual(form.initial['internal_type'], Organization.AFFILIATED)

        form = AffiliatedOrganizationForm(initial={'internal_type': Organization.NORMAL})
        self.assertEqual(form.initial['internal_type'], Organization.AFFILIATED)

    def test_clean_internal_type(self):
        form_data = {
            'internal_type': Organization.NORMAL,
            'data_source': self.data_source.id,
            'origin_id': 'sub-org',
            'classification': self.organization_class.id,
            'name': 'sub organization',
        }
        form = AffiliatedOrganizationForm(form_data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['internal_type'], Organization.AFFILIATED)

        form_data['internal_type'] = Organization.AFFILIATED
        form = AffiliatedOrganizationForm(form_data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['internal_type'], Organization.AFFILIATED)



# Generated by Django 2.2.8 on 2020-01-08 08:44

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_orghierarchy', '0007_set_editable_false_on_mptt_fields'),
    ]

    operations = [
        migrations.AlterField(
            model_name='organization',
            name='created_by',
            field=models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='created_organizations', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='organization',
            name='last_modified_by',
            field=models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='modified_organizations', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='organization',
            name='replaced_by',
            field=models.OneToOneField(blank=True, help_text='The organization that replaces this organization', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='replaced_organization', to='django_orghierarchy.Organization'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-08-27 13:31
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_orghierarchy', '0005_add_fields_to_organization_class'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='organization',
            options={'permissions': (('add_affiliated_organization', 'Can add affiliated organization'), ('change_affiliated_organization', 'Can change affiliated organization'), ('delete_affiliated_organization', 'Can delete affiliated organization'), ('replace_organization', 'Can replace an organization with a new one'))},
        ),
        migrations.AddField(
            model_name='datasource',
            name='user_editable',
            field=models.BooleanField(default=False, verbose_name='Objects may be edited by users'),
        ),
    ]



from django.conf.urls import include, url
from django.contrib import admin
from rest_framework.routers import DefaultRouter

from django_orghierarchy.api import OrganizationViewSet

router = DefaultRouter()
router.register(r'organization', OrganizationViewSet, 'organization')

urlpatterns = [
    url(r"^admin/", admin.site.urls),
    url(r'^api/', include((router.urls, 'api'), namespace='api')),
]



import swapper
from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from mptt.models import MPTTModel, TreeForeignKey


class AbstractDataSource(models.Model):
    """Abstract data source model.

    Abstract data source model that provides required fields
    for custom data source model.
    """
    id = models.CharField(max_length=100, primary_key=True)
    name = models.CharField(max_length=255)
    user_editable_organizations = models.BooleanField(default=False, verbose_name=_('Organizations may be edited by users'))

    class Meta:
        abstract = True

    def __str__(self):
        return self.name


class DataSource(AbstractDataSource):
    """Default data source model.

    The default data source model will be used if the project
    does not provide a custom data source model.
    """

    class Meta:
        swappable = swapper.swappable_setting('django_orghierarchy', 'DataSource')


class DataModel(models.Model):
    id = models.CharField(max_length=255, primary_key=True, editable=False)
    data_source = models.ForeignKey(
        swapper.get_model_name('django_orghierarchy', 'DataSource'), on_delete=models.CASCADE, blank=True, null=True)
    origin_id = models.CharField(max_length=255, blank=True)
    created_time = models.DateTimeField(default=timezone.now, help_text=_('The time at which the resource was created'))
    last_modified_time = models.DateTimeField(auto_now=True, help_text=_('The time at which the resource was updated'))

    class Meta:
        abstract = True
        unique_together = ('data_source', 'origin_id')

    def save(self, *args, **kwargs):
        if not self.id:
            # the id is only set when creating object, it cannot be changed later
            self.id = '{0}:{1}'.format(self.data_source_id, self.origin_id)
        super().save(*args, **kwargs)


class OrganizationClass(DataModel):
    name = models.CharField(max_length=255)

    class Meta:
        unique_together = ('data_source', 'origin_id')
        verbose_name = _('Organization class')
        verbose_name_plural = _('Organization classes')

    def __str__(self):
        return self.name


class Organization(MPTTModel, DataModel):
    NORMAL = 'normal'
    AFFILIATED = 'affiliated'

    INTERNAL_TYPES = (
        (NORMAL, _('Normal organization')),
        (AFFILIATED, _('Affiliated organization')),
    )

    internal_type = models.CharField(max_length=20, choices=INTERNAL_TYPES, default=NORMAL)

    classification = models.ForeignKey(OrganizationClass, on_delete=models.PROTECT, blank=True, null=True,
                                       help_text=_('An organization category, e.g. committee'))
    name = models.CharField(max_length=255, help_text=_('A primary name, e.g. a legally recognized name'))
    founding_date = models.DateField(blank=True, null=True, help_text=_('A date of founding'))
    dissolution_date = models.DateField(blank=True, null=True, help_text=_('A date of dissolution'))
    parent = TreeForeignKey('self', on_delete=models.CASCADE, null=True, blank=True, related_name='children',
                            help_text=_('The organizations that contain this organization'))
    admin_users = models.ManyToManyField(settings.AUTH_USER_MODEL, blank=True, related_name='admin_organizations')
    regular_users = models.ManyToManyField(settings.AUTH_USER_MODEL, blank=True,
                                           related_name='organization_memberships')
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, related_name='created_organizations',
        null=True, blank=True, editable=False)
    last_modified_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, related_name='modified_organizations',
        null=True, blank=True, editable=False)
    replaced_by = models.OneToOneField(
        'self', on_delete=models.SET_NULL, null=True, blank=True, related_name='replaced_organization',
        help_text=_('The organization that replaces this organization'))

    class Meta:
        unique_together = ('data_source', 'origin_id')
        permissions = (
            ('add_affiliated_organization', 'Can add affiliated organization'),
            ('change_affiliated_organization', 'Can change affiliated organization'),
            ('delete_affiliated_organization', 'Can delete affiliated organization'),
            ('replace_organization', 'Can replace an organization with a new one'),
            ('change_organization_regular_users', 'Can add/remove regular users to organizations'),
        )

    def __str__(self):
        if self.dissolution_date:
            return self.name + ' (dissolved)'
        return self.name

    @property
    def sub_organizations(self):
        return self.children.filter(internal_type=self.NORMAL)

    @property
    def affiliated_organizations(self):
        return self.children.filter(internal_type=self.AFFILIATED)



import os
from setuptools import find_packages, setup

from django_orghierarchy import __version__

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-orghierarchy',
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='Reusable Django application for hierarchical organizations.',
    long_description=README,
    url='https://github.com/City-of-Helsinki/django-orghierarchy',
    author='City of Helsinki',
    author_email='dev@hel.fi',
    install_requires=['django-mptt', 'djangorestframework', 'requests', 'swapper'],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-22 15:58
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CustomDataSource',
            fields=[
                ('id', models.CharField(max_length=100, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255)),
                ('extra_field', models.CharField(max_length=200)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='CustomPrimaryKeyDataSource',
            fields=[
                ('id', models.CharField(max_length=100, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255)),
            ],
        ),
    ]



# Generated by Django 2.2.8 on 2020-02-03 12:04

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_orghierarchy', '0008_modify_organization_on_deletes'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='organization',
            options={'permissions': (('add_affiliated_organization', 'Can add affiliated organization'), ('change_affiliated_organization', 'Can change affiliated organization'), ('delete_affiliated_organization', 'Can delete affiliated organization'), ('replace_organization', 'Can replace an organization with a new one'), ('change_organization_regular_users', 'Can add/remove regular users to organizations'))},
        ),
    ]



from django.contrib import admin
from django.test import TestCase, tag

from django_orghierarchy.utils import get_data_source_model
from django_orghierarchy.models import DataSource, Organization
from ..models import CustomDataSource


@tag('custom_ds')
class TestCustomDataSource(TestCase):

    def test_get_data_source_model(self):
        model = get_data_source_model()
        self.assertIs(model, CustomDataSource)

    def test_related_data_source_model(self):
        field = Organization._meta.get_field('data_source')
        self.assertIs(field.related_model, CustomDataSource)

    def test_default_data_source_admin_not_registered(self):
        is_registered = admin.site.is_registered(DataSource)
        self.assertFalse(is_registered)



from django.test import TestCase

from django_orghierarchy.models import Organization
from .factories import DataSourceFactory, OrganizationClassFactory, OrganizationFactory


class TestDataSource(TestCase):

    def setUp(self):
        self.data_source = DataSourceFactory(name='test name')

    def test__str__(self):
        self.assertEqual(self.data_source.__str__(), 'test name')


class TestOrganizationClass(TestCase):

    def setUp(self):
        self.organization_class = OrganizationClassFactory(name='test name')

    def test__str__(self):
        self.assertEqual(self.organization_class.__str__(), 'test name')


class TestOrganization(TestCase):

    def setUp(self):
        self.parent_organization = OrganizationFactory(name='parent name')
        self.organization = OrganizationFactory(name='test name', parent=self.parent_organization)
        self.affiliated_organization = OrganizationFactory(
            name='test aff org',
            parent=self.parent_organization,
            internal_type=Organization.AFFILIATED,
        )

    def test__str__(self):
        self.assertEqual(self.parent_organization.__str__(), 'parent name')
        self.assertEqual(self.organization.__str__(), 'test name')

    def test_save(self):
        data_source = DataSourceFactory(id='data-source')
        organization = OrganizationFactory(data_source=data_source, origin_id='ABC123')
        self.assertEqual(organization.id, 'data-source:ABC123')

        organization.origin_id = 'XYZ'
        organization.save()
        # test the id is not changed
        self.assertEqual(organization.id, 'data-source:ABC123')

    def test_sub_organizations(self):
        qs = self.parent_organization.sub_organizations
        self.assertQuerysetEqual(qs, [repr(self.organization)])

    def test_affiliated_organizations(self):
        qs = self.parent_organization.affiliated_organizations
        self.assertQuerysetEqual(qs, [repr(self.affiliated_organization)])
